from typing import Generator, Any, Optional, Tuple

# ! Base Class
class ErrorBase(Exception):
    """ """
    def __init__(self, *args, **kwargs) -> None:
        """ """
        self.__init__()
        self.args: Tuple[str, ...] = tuple([arg for arg in self.exception(*args, **kwargs) if arg is not None])
    
    def exception(self, *args, **kwargs) -> Generator[Optional[str], Any, None]:
        yield None

# ! ChunkFile Errors
class ChunkIsDamagedError(ErrorBase):
    def exception(self, file_crc: int, chunk_crc: int, *args, **kwargs):
        yield f"The checksum of the read chunk does not match the checksum of the one written in the chunk file ({repr(file_crc)} != {repr(chunk_crc)})."

class IsNotChunkFile(ErrorBase):
    def exception(self, *args, **kwargs):
        yield "The signature of the file is not equal to the signature of the specified chunk file."

# ! IO Errors
class IOReadOnlyError(ErrorBase):
    def exception(self, *args, **kwargs):
        yield "This IO is open in read-only mode."


import os
from pathlib import Path
from .crc import CRC32

# ! Functions
def initfp(fp):
    if isinstance(fp, str) or isinstance(fp, Path):
        path = str(Path(fp))
        if os.path.exists(path):
            assert os.path.isfile(path)
            return str(path), open(path, 'rb+')
        else:
            return str(path), open(path, 'wb+')
    else:
        assert not fp.closed
        assert fp.readable()
        assert fp.seekable()
        assert fp.writable()
        assert hasattr(fp, "name")
        assert isinstance(fp.name, str)
        return fp.name, fp

def iocopy(from_io, to_io, size, buffer_size=8192):
    crc = CRC32()
    while size > 0:
        if size >= buffer_size:
            rsize, size = buffer_size, size - buffer_size
        else:
            rsize, size = size, 0
        d: bytes = from_io.read(rsize)
        crc.update(d)
        to_io.write(d)
    return crc


import zlib

# ! Main Class
class CRC32:
    def __init__(self, start_value: int=0) -> None:
        self.__crc = start_value
    
    # ! Property
    @property
    def crc(self) -> int: return self.__crc
    @property
    def crc_hex(self) -> int: return hex(self.__crc)
    @property
    def crc_bytes(self) -> bytes: return self.__crc.to_bytes(4, "big")
    
    # ! Magic Methods
    def __str__(self) -> str: return f"{self.__class__.__name__}(crc={repr(self.crc)}, crc_hex={repr(self.crc_hex)}, crc_bytes={repr(self.crc_bytes)})"
    def __repr__(self) -> str: return self.__str__()
    
    # ! CRC32 Methods
    def update(self, data: bytes) -> None:
        self.__crc = zlib.crc32(data, self.__crc)
